#include <ansi_c.h>
#include <analysis.h>
//==============================================================================
//
// Title:       ACF
// Purpose:     A short description of the library.
//
// Created on:  2008/4/20 at U 07:21:15 by Hsu-Yang Lee.
// Copyright:   NTU. All Rights Reserved.
//
//==============================================================================

//==============================================================================
// Include files

#include <utility.h>
#include "ACF.h"

//==============================================================================
// Constants
#define UNITCONVERT_S2NS 1e9
#define FAILURE 0
#define VERSION_REALACF 20070817

//==============================================================================
// Types

//==============================================================================
// Static global variables

//==============================================================================
// Static functions

//==============================================================================
// Global variables
struct ACF 
{
	long *delay; /* buffer: delay interphoton times */
	long mdelay; /* buffer size: maximum number of interphoton times to keep in memory*/
	long idelay; /* current position in buffer */
	double dutyFactor; /* maximum duty factor (0-1) per clock period */
	long *acf; /* buffer: auto correlation */
	long mtau; /* buffer size: acf function */
	double *nacf; /* pointer to normalized ACF function array */
	double *dblacf;
	double *dblacffit;
	double *dblacfx;
	double fitcoef[4];
	unsigned __int64 acnt, tcnt; /*counts of ipt arr and total timebase pulse, for normalization*/
	
} d;


//==============================================================================
// Global functions


double __stdcall calcRealACFdutyFactor(double m,long c){return( m*c/UNITCONVERT_S2NS);}

long __stdcall calcRealACFmtau(long m, long c){return(m/c);}

double ACFFitFunc(double x, double a[], int ncoef)
{
	return(a[0]/(1+x/a[1])/pow((1+x*a[2]/a[1]),0.5)+a[3]);
}


long __stdcall setupRealACF(long mtau,double dutyFactor, double clock_ns)
{
	long i=0;
	d.mtau=mtau;
	d.dutyFactor=dutyFactor;
	d.idelay=0;
	d.mdelay=(long) (dutyFactor*mtau+1);
	d.acf=(long *)calloc(mtau+1,sizeof(long));
	d.delay=(long *)calloc(d.mdelay,sizeof(long));
	d.dblacf=(double *)calloc(mtau+1,sizeof(double));
	d.dblacffit=(double *)calloc(mtau+1,sizeof(double));
	d.dblacfx=(double *)calloc(mtau+1,sizeof(double));
	d.nacf=(double *)calloc(mtau+1,sizeof(double));
	for (i=0;i<mtau+1;i++)
		d.dblacfx[i]=i*clock_ns;
	if(d.acf==NULL || d.delay==NULL)goto error;
	d.acf[0]=1;
	d.acnt=0;
	d.tcnt=0;
	return(d.mdelay);
error:
	printf("setupRealACF: ERROR");
	if(d.acf==NULL)
		printf("memory allocation failed for 'acf' w/mtau=%ld",(long) d.mtau);
	else if(d.delay==NULL)
		printf("memory allocation failed for 'delay' w/mtau=%ld, dutyFactor=%f",
			(long) d.mtau,(double) d.dutyFactor);
	return(FAILURE);
}

long __stdcall realACFraw(long tau)
{
	long i=0;
	
	d.acnt++;
	d.tcnt+=tau;
	
	if(d.idelay > d.mdelay)goto error;
	while(d.delay[d.idelay]+tau > d.mtau && d.idelay>0){
		d.delay[d.idelay]=0;
		d.idelay--;
	}
	d.idelay++;
	if(tau>d.mtau)d.idelay=0;
	for(i=d.idelay;i>0;i--)d.delay[i]=d.delay[i-1]+tau;
	for(i=0;i<=d.idelay;i++)d.acf[d.delay[i]]++;
	return(tau);
error: 
	printf("realACF : ERROR count rate too high for time delay.\n\t"
		"\n\t SUGGESTION: increase dutyFactor by a factor of %.2f or decrease excitation power",
		(double) d.mtau /d.delay[d.idelay-1]);
	return(FAILURE);
}

void __stdcall realACFnorm(char *dst)
{
	long i=0;
	FILE *fdst = fopen (dst, "wb");
	
	double dblacnt=(double)d.acnt, dbltcnt=(double)d.tcnt;
	
	for(i=0;i<d.mtau+1;i++)d.nacf[i]=d.acf[i]/pow(dblacnt,2)*pow(dbltcnt,2)/(dbltcnt-i)-1.0;
	fwrite(d.nacf, sizeof(double), d.mtau+1, fdst);
	fclose(fdst);
	
	return;
	
}

double __stdcall  fitRealACF(long maxiter)
{
	double fiterror;
	double midpt;
	long uidx=d.mtau, lidx=0, i=0;
	
	if (d.mtau>10)
	{
		for(i=0;i<d.mtau+1;i++)d.dblacf[i]=(double)d.acf[i];
	
		midpt=0.25*d.dblacf[3]+0.75*d.dblacf[d.mtau-3];
		for(i=0;i<20;i++)
		{
			if(d.dblacf[(long)((uidx+lidx)/2.0)]>midpt)
				lidx=(long)((uidx+lidx)/2.0);
			else
				uidx=(long)((uidx+lidx)/2.0); 
		}
	
		d.fitcoef[1]=d.dblacfx[(long)((uidx+lidx)/2.0)];

		d.fitcoef[3]=d.dblacf[3]; 	
		d.fitcoef[0]=d.dblacf[d.mtau-3]-d.fitcoef[3];
		d.fitcoef[2]=3.0;
		
		NonLinearFitWithMaxIters (d.dblacfx, d.dblacf, d.dblacffit, d.mtau, maxiter, ACFFitFunc, d.fitcoef, 4, &fiterror);				   
	}
	return fiterror;
		
}

void _stdcall getRealACF(long *acfarr)
{
	long i=0;
	for(i=0;i<d.mtau+1;i++)
		acfarr[i]=d.acf[i];
}

void _stdcall getRealACFnorm(double *acfnorm)
{
	long i=0;
	for(i=0;i<d.mtau+1;i++)
		acfnorm[i]=d.nacf[i];
}

void _stdcall getRealACFfit(double *dblacffit, double *acfx)
{
	long i=0;
	for(i=0;i<d.mtau+1;i++)
	{
		dblacffit[i]=d.dblacffit[i];
		acfx[i]=d.dblacfx[i];
	}
}

void __stdcall freeRealACF(void)
{
	free(d.acf);
	free(d.delay);
	free(d.nacf);
	free(d.dblacffit);
	free(d.dblacf);
	free(d.dblacfx);
}

void __stdcall file2acf(char *cfn, char *dfn, long mtau, double dutyFactor, double clock_ns)
{
	long i=0, err=0;
	long lngipt=0;
	FILE* fsrc=fopen(cfn, "rb");
	FILE *fdes;
	
	setupRealACF(mtau, dutyFactor, clock_ns);
	while(feof(fsrc)==0)
	{
		err=fread (&lngipt, sizeof(long), 1, fsrc);
		realACFraw(lngipt);

	}
	fclose(fsrc);
	fdes = fopen (dfn, "wb");
	if(fdes==NULL)goto error;					
	fwrite(d.acf,sizeof(long),d.mtau+1 ,fdes);
	//for(i=0;i<d.mtau+1;i++)fprintf(fdes,"%f\t%d\n",d.dblacfx[i] ,d.acf[i]);
	fclose(fdes);
	return;
error:
	return;
	
}

void __stdcall getACFcnt(unsigned __int64 *acntpt, unsigned __int64 *tcntpt)
{
	(*acntpt)=d.acnt;
	(*tcntpt)=d.tcnt;
	return;
}

void __stdcall resetrealACF(void){d.idelay=0;}


//==============================================================================
// DLL main entry-point functions

int __stdcall DllMain (HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
    switch (fdwReason) {
        case DLL_PROCESS_ATTACH:
            if (InitCVIRTE (hinstDLL, 0, 0) == 0)
                return 0;     /* out of memory */
            break;
        case DLL_PROCESS_DETACH:
            CloseCVIRTE ();
            break;
    }
    
    return 1;
}

int __stdcall DllEntryPoint (HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
    /* Included for compatibility with Borland */

    return DllMain (hinstDLL, fdwReason, lpvReserved);
}
